/*-
 * Copyright (C)2003..2013 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: mife.h,v 1.21 2013/10/13 19:23:14 babolo Exp $
 */

#ifndef  _BABOLO_BLINflag_H_
# warning this file needs babolo/BLINflag.h as a prerequisite
# include <babolo/BLINflag.h>
#endif

#ifndef _MIFE_H_
#define _MIFE_H_

__BEGIN_DECLS

#define MIFE_RTRY    0x0000FF /* R  */
#define MIFE_BUFL    0x000F00 /* RW */
#define MIFE_SLID    0x001000 /* R         */
#define MIFE_RECY    0x001000 /*  W    */
#define MIFE_STRI    0x002000 /* R  not use strlen(src) in CTLCNST & CTLBUFF */
#define MIFE_SYNC    0x002000 /*  W   AIO            */
#define MIFE_INMM    0x010000 /* R               */
#define MIFE_ALLC    0x020000 /* RW    free()        */
#define MIFE_MMAP    0x040000 /* R   mmaped                    */
#define MIFE_CLOS    0x080000 /* RW    fd           */
#define MIFE_EOFL    0x100000 /* R               */
                              /*  W   EOF     */
#define MIFE_FULL    0x200000 /* R    \0            */
#define MIFE_PIPE    0x400000 /* R         */
#define MIFE_BUFX    0x800000 /* R          */

#define MIFE_CTLFLAG 0x000000
#define MIFE_CTLFILE 0x000001
#define MIFE_CTLFDSC 0x000002
#define MIFE_CTLCNST 0x000003
#define MIFE_CTLBUFF 0x000004
#define MIFE_CTLEOLI 0x000005
#define MIFE_CTLITIM 0x000006
#define MIFE_CTLOTIM 0x000007

#define kukMIFE_RTRY(x)    ((u_int32_t)(x) & MIFE_RTRY)
#define setMIFE_RTRY(a, x) do {(a) &= ~MIFE_RTRY; (a) |= kukMIFE_RTRY(x);} while(0)
#define getMIFE_RTRY(a)    ((a) & MIFE_RTRY)
#define kukMIFE_BUFL(x)    (((u_int32_t)(x) << 8) & MIFE_BUFL)
#define setMIFE_BUFL(a, x) do {(a) &= ~MIFE_BUFL; (a) |= kukMIFE_BUFL(x);} while(0)
#define getMIFE_BUFL(a)    (((a) >> 8) & (MIFE_BUFL >> 8))

typedef struct timespec mife_time;
struct mife_descriptor;
typedef struct mife_descriptor mife_descriptor;

typedef struct mife_eol {
    ssize_t   (*init)(mife_descriptor *mife, const char *eoln);
                         /*   eol           */
    ssize_t   (*detect)(mife_descriptor *mife);
                         /*                               */
    void      (*start)(mife_descriptor *mife, ssize_t delta);
    void      (*delta)(mife_descriptor *mife, ssize_t delta);
    void      (*stop)(mife_descriptor *mife);
    const char *ln;      /*                       */
    ssize_t    *trix;    /*               */
    size_t      max;     /*                          */
    size_t      trilen;  /*                    */
    ssize_t     p;       /*          */
    ssize_t     st;      /*     */
} mife_eol;

typedef struct mife_event {
    int           kq;
    struct kevent kev;
} mife_event;

struct mife_descriptor {
    BLIN_flag   flags;   /*                                  */
    mode_t      st_mode; /*                              */
    int16_t     dummy;
    off_t       st_size; /*   ( )              */
    off_t       offset;  /*            */
    off_t       ofpoint; /*                   */
    mife_eol   *eol;     /*                  */
    void       *buffer;  /*                                  */
    mife_event *kev;     /*  kevent     */
    size_t      buflen;  /*                           */
    size_t      actbeg;  /*              */
    size_t      actlen;  /*              */
    ssize_t     mindelta;/*         */
    mife_time   itime;   /*              */
    mife_time   otime;   /*      */
    union {
        struct {
            int     (*open)(void *, ...);
            ssize_t (*read)(int f, void *buf, size_t nbytes);
            int     (*close)(int f);
            int       fd;/*               */
        } f;
        struct {
            void   *(*open)(void *, ...);
            ssize_t (*read)(void *a, void *buf, size_t nbytes);
            int     (*close)(void *a);
            void     *ad;/*         */
        } a;
    };
};

/*      64 */
typedef struct mife_aout {
    BLIN_flag   flag;    /*                                  */
    union { size_t        pos;
            struct {
                u_int16_t lo;
                u_int16_t hi;
            } wr;
    } towrite;           /*                */
    off_t       offset;  /*            */
    void       *iocb[];  /*         */
} mife_aout;

extern BLIN_flag mife_default;
extern mife_eol mife_eol0;

mife_descriptor *mife_init      __P((BLIN_flag flags));
int              mife_ctl       __P((mife_descriptor *mife, int type, ...));
ssize_t          mife_read      __P((mife_descriptor *mife, ssize_t ordlen, off_t offset));
void *           mife_get       __P((mife_descriptor *mife, off_t offset));
int              mife_fini      __P((mife_descriptor *mife));

ssize_t          mife_writ      __P((int fd, const void *buf, size_t nbytes));
mife_aout *      mife_aopen     __P((BLIN_flag flags, int fd));
int              mife_aput      __P((mife_aout *aso, int src));
int              mife_aclose    __P((mife_aout *aso));

#ifdef MIFE_COMPAT3

mife_descriptor *mife_opef      __P((BLIN_flag flags, int fd));
mife_descriptor *mife_open      __P((BLIN_flag flags, const char *path));
mife_descriptor *mife_opes      __P((BLIN_flag flags, const char *src, size_t size));
mife_descriptor *mife_opem      __P((BLIN_flag flags, char *src, size_t size));

ssize_t          mife_eolinit   __P((mife_descriptor *mife, const char *eoln));
ssize_t          mife_eoldetect __P((mife_descriptor *mife));

#define eomax      eol->max
#define eotrilen   eol->trilen
#define eolp       eol->p
#define eolst      eol->st
#define eotrix     eol->trix
#define eoln       eol->ln
#define mindelta   type->mindelta
#define mife_close mife_fini

#endif

__END_DECLS

#endif /* !_MIFE_H_ */
